from typing import Dict, Union
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from spark_expectations import _log
from spark_expectations.notifications.plugins.base_notification import (
    SparkExpectationsNotification,
    spark_expectations_notification_impl,
)
from spark_expectations.core.exceptions import SparkExpectationsEmailException
from spark_expectations.core.context import SparkExpectationsContext


# Create the email plugin
class SparkExpectationsEmailPluginImpl(SparkExpectationsNotification):
    """
    This class implements/supports functionality to send email
    """

    @spark_expectations_notification_impl
    def send_notification(
        self,
        _context: SparkExpectationsContext,
        _config_args: Dict[Union[str], Union[str, bool]],
    ) -> None:
        """
        function to send email notification for requested mail id's
        Args:
            _context: object of SparkExpectationsContext
            _config_args: dict(which consist to: receiver mail(str), subject: subject of
                          the mail(str) and body: body of the mail(str)
        Returns:

        """
        try:
            if _context.get_enable_mail is True:
                msg = MIMEMultipart()
                msg["From"] = _context.get_mail_from
                msg["To"] = _context.get_to_mail
                msg["Subject"] = _context.get_mail_subject

                # body = _config_args.get('mail_body')

                mail_content = f"""{_config_args.get("message")}"""
                msg.attach(MIMEText(mail_content, "plain"))

                # mailhost.com
                server = smtplib.SMTP(
                    _context.get_mail_smtp_server, _context.get_mail_smtp_port
                )
                server.starttls()
                text = msg.as_string()
                server.sendmail(_context.get_mail_from, _context.get_to_mail, text)
                server.quit()

                _log.info("email send successfully")

        except Exception as e:
            raise SparkExpectationsEmailException(
                f"error occurred while sending email notification from spark expectations project {e}"
            )



import sys
import logging
from typing import Optional


def get_default_log_handler(pkg_name: str = "spark_expectations") -> logging.Handler:
    logger_handler = logging.StreamHandler(stream=sys.stdout)
    logger_handler.setFormatter(
        logging.Formatter(
            f"[%(asctime)s] [%(levelname)s] [{pkg_name}] "
            "{%(module)s.py:%(funcName)s:%(lineno)d} - %(message)s"
        )
    )
    return logger_handler


def setup_logger(name: Optional[str] = None) -> logging.Logger:
    log = logging.getLogger(name or __name__)  # Logger
    log.setLevel(logging.INFO)
    logger_handler = get_default_log_handler()
    log.propagate = False
    if log.hasHandlers():
        log.handlers.clear()
    log.addHandler(logger_handler)
    return log


_log = setup_logger()



import os
import pandas as pd
import numpy as np
import pkg_resources
from pyspark.sql import DataFrame
from spark_expectations.core import get_spark_session

spark = get_spark_session()


def spark_expectations_read_employee_dataset() -> DataFrame:
    _df: DataFrame = pd.read_csv(
        pkg_resources.resource_filename(
            "spark_expectations", "examples/resources/employee.csv"
        )
    ).replace({np.nan: None})

    return spark.createDataFrame(_df)


def spark_expectations_read_order_dataset() -> DataFrame:
    _df: DataFrame = (
        spark.read.option("header", "true")
        .option("inferSchema", "true")
        .csv("file://" + os.path.join(os.path.dirname(__file__), "resources/order.csv"))
    )
    return _df


def spark_expectations_read_product_dataset() -> DataFrame:
    _df: DataFrame = (
        spark.read.option("header", "true")
        .option("inferSchema", "true")
        .csv(
            "file://" + os.path.join(os.path.dirname(__file__), "resources/product.csv")
        )
    )
    return _df


def spark_expectations_read_customer_dataset() -> DataFrame:
    _df: DataFrame = (
        spark.read.option("header", "true")
        .option("inferSchema", "true")
        .csv(
            "file://"
            + os.path.join(os.path.dirname(__file__), "resources/customer.csv")
        )
    )
    return _df



# from pyspark.sql.utils import CapturedException


class SparkExpectationsDataframeNotReturnedException(Exception):
    """
    Throw this exception if a function doesn't return a dataframe
    """

    pass


class SparkExpectOrFailException(Exception):
    """
    Throw this exception if a rule fails and is expected to fail the job
    """

    pass


class SparkExpectationsUserInputOrConfigInvalidException(Exception):
    """
    Throw this exception when configured rule or value from the user is wrong
    """

    pass


# class SparkExpectationsSqlQueryParserException(CapturedException):
#     """
#     Throw this exception when spark expectations fail to parse the given user query
#     """
#
#     pass


class SparkExpectationsMiscException(Exception):
    """
    Throw this exception when spark expectations encounters miscellaneous exceptions
    """

    pass


class SparkExpectationsSlackNotificationException(Exception):
    """
    Throw this exception when spark expectations encounters exceptions while sending Slack notifications
    """

    pass


class SparkExpectationsTeamsNotificationException(Exception):
    """
    Throw this exception when spark expectations encounters exceptions while sending Teams notifications
    """

    pass


class SparkExpectationsEmailException(Exception):
    """
    Throw this exception when spark expectations encounters exceptions while sending email notifications
    """

    pass


class SparkExpectationsErrorThresholdExceedsException(Exception):
    """
    Throw this exception when error percentage exceeds certain configured value
    """

    pass



from typing import Dict, Union
from pyspark.sql import DataFrame
import pluggy

SPARK_EXPECTATIONS_WRITER_PLUGIN = "spark_expectations_writer_plugins"

writer_plugin_spec = pluggy.HookspecMarker(SPARK_EXPECTATIONS_WRITER_PLUGIN)
spark_expectations_writer_impl = pluggy.HookimplMarker(SPARK_EXPECTATIONS_WRITER_PLUGIN)


class SparkExpectationsSinkWriter:
    @writer_plugin_spec
    def writer(
        self, _write_args: Dict[Union[str], Union[str, bool, Dict[str, str], DataFrame]]
    ) -> None:
        """
        function consist signature to write data into kafka etc. which will be implemented in the child class
        Args:
            _write_args:

        Returns:

        """

        pass



from typing import Dict, Union
import requests
from spark_expectations import _log
from spark_expectations.notifications.plugins.base_notification import (
    SparkExpectationsNotification,
    spark_expectations_notification_impl,
)
from spark_expectations.core.exceptions import (
    SparkExpectationsTeamsNotificationException,
)
from spark_expectations.core.context import SparkExpectationsContext


class SparkExpectationsTeamsPluginImpl(SparkExpectationsNotification):
    """
    This class implements/supports functionality to send teams notification
    """

    @spark_expectations_notification_impl
    def send_notification(
        self,
        _context: SparkExpectationsContext,
        _config_args: Dict[Union[str], Union[str, bool]],
    ) -> None:
        """
        function to send the teams notification for assigned channel
        Args:
            _context: SparkExpectationsContext class object
            _config_args: dict

        Returns: None

        """
        try:
            if _context.get_enable_teams is True:
                # payload = {"token": "{token}", "channel": kwargs['channel'], "text": kwargs['message']}

                message = _config_args.get("message")

                # Format Message for Teams
                if isinstance(message, str):
                    message = message.replace("\n", "\n\n").replace("            ", "")

                payload = {
                    "title": "SE Notification",
                    "themeColor": "008000",
                    "text": message,
                }

                response = requests.post(
                    _context.get_teams_webhook_url, json=payload, timeout=10
                )

                # Check the response for success or failure
                if response.status_code == 200:
                    _log.info("Message posted successfully!")
                else:
                    _log.info("Failed to post message")
                    raise SparkExpectationsTeamsNotificationException(
                        "error occurred while sending teams notification from spark expectations project"
                    )

        except Exception as e:
            raise SparkExpectationsTeamsNotificationException(e)



from typing import Dict, Union
import pluggy
from spark_expectations.core.context import SparkExpectationsContext

SPARK_EXPECTATIONS_NOTIFICATION_PLUGIN = "spark_expectations_notification_plugins"

notification_plugin_spec = pluggy.HookspecMarker(SPARK_EXPECTATIONS_NOTIFICATION_PLUGIN)
spark_expectations_notification_impl = pluggy.HookimplMarker(
    SPARK_EXPECTATIONS_NOTIFICATION_PLUGIN
)


class SparkExpectationsNotification:
    """
    This is base class for notifications plugin
    """

    @notification_plugin_spec
    def send_notification(
        self,
        _context: SparkExpectationsContext,
        _config_args: Dict[Union[str], Union[str, bool]],
    ) -> None:
        """
        function consist signature to notification, which will be implemented in the child class
        Args:
            _context:object of SparkExpectationsContext
            _config_args: dict which contains required parameter to send notification

        Returns: None

        """
        pass



import functools
import pluggy
from spark_expectations import _log
from spark_expectations.sinks.plugins.base_writer import (
    SparkExpectationsSinkWriter,
    SPARK_EXPECTATIONS_WRITER_PLUGIN,
)

from spark_expectations.sinks.plugins.kafka_writer import (
    SparkExpectationsKafkaWritePluginImpl,
)


@functools.lru_cache
def get_sink_hook() -> pluggy.PluginManager:
    """
    function provides pluggy hook manager to write data into delta and nsp
    Returns:
        PluginManager: pluggy Manager object

    """
    pm = pluggy.PluginManager(SPARK_EXPECTATIONS_WRITER_PLUGIN)
    pm.add_hookspecs(SparkExpectationsSinkWriter)
    pm.register(
        SparkExpectationsKafkaWritePluginImpl(), "spark_expectations_kafka_write"
    )
    for name, plugin_instance in pm.list_name_plugin():
        _log.info(
            "Loaded plugin with name: %s and class: %s",
            name,
            plugin_instance.__class__.__name__,
        )
    return pm


_sink_hook = get_sink_hook().hook



from dataclasses import dataclass
from typing import Any, List, Dict
from pyspark.sql import DataFrame
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.notifications import _notification_hook
from spark_expectations.core.exceptions import SparkExpectationsMiscException


@dataclass
class SparkExpectationsNotify:
    """
    This class implements Notification
    """

    _context: SparkExpectationsContext

    def __post_init__(self) -> None:
        self.send_notification_decorator: Any = self.notify_on_start_completion_failure(
            self.notify_on_start, self.notify_on_completion, self.notify_on_failure
        )

    def notify_on_start_completion_failure(
        self, _on_start: Any, _on_completion: Any, _on_failure: Any
    ) -> Any:
        """
        This function orchestrate notification
        Args:
            _on_start: function to send notification on start of spark expectations
            _on_completion: function to send notification on completion of spark expectations
            _on_failure: function to send notification on failure

        Returns: decorated notification function

        """

        def decorator(func: Any) -> Any:
            def wrapper(*args: List, **kwargs: Dict) -> DataFrame:
                if self._context.get_notification_on_start is True:
                    _on_start()
                try:
                    # self._context.set_dq_start_time()
                    result = func(*args, **kwargs)

                    if self._context.get_notification_on_completion is True:
                        _on_completion()
                    # self._context.set_dq_end_time()

                except Exception as e:
                    # self._context.set_dq_run_status("Failed")
                    if self._context.get_notification_on_fail is True:
                        _on_failure(e)
                    # self._context.set_dq_end_time()
                    raise SparkExpectationsMiscException(e)
                return result

            return wrapper

        return decorator

    def notify_on_start(self) -> None:
        """
        This function sends notification on start of spark expectations project
        Returns: None

        """

        _notification_message = (
            "Spark expectations job has started \n\n"
            f"table_name: {self._context.get_table_name}\n"
            f"run_id: {self._context.get_run_id}\n"
            f"run_date: {self._context.get_run_date}"
        )

        _notification_hook.send_notification(
            _context=self._context,
            _config_args={
                "message": _notification_message,
            },
        )

    def notify_on_exceeds_of_error_threshold(self) -> None:
        """
        This function sends notification on completion of spark expectations project
        Returns: None

        """

        _notification_message = (
            f"Spark expectations - dropped error percentage has been exceeded above the threshold "
            f"value({self._context.get_error_drop_threshold}%) for `row_data` quality validation  \n\n"
            f"product_id: {self._context.product_id}\n"
            f"table_name: {self._context.get_table_name}\n"
            f"run_id: {self._context.get_run_id}\n"
            f"run_date: {self._context.get_run_date}\n"
            f"input_count: {self._context.get_input_count}\n"
            f"error_percentage: {self._context.get_error_percentage}\n"
            f"error_drop_percentage: {self._context.get_error_drop_percentage}\n"
            f"output_percentage: {self._context.get_output_percentage}\n"
            f"success_percentage: {self._context.get_success_percentage}"
            # f"status: source_agg_dq_status = {self._context.get_source_agg_dq_status}\n"
            # f"            source_query_dq_status = {self._context.get_source_query_dq_status}\n"
            # f"            row_dq_status = {self._context.get_row_dq_status}\n"
            # f"            final_agg_dq_status = {self._context.get_final_agg_dq_status}\n"
            # f"            final_query_dq_status = {self._context.get_final_query_dq_status}\n"
            # f"            run_status = {self._context.get_dq_run_status}"
        )

        _notification_hook.send_notification(
            _context=self._context, _config_args={"message": _notification_message}
        )

    def notify_on_completion(self) -> None:
        """
        This function sends notification on completion of spark expectations project
        Returns: None

        """

        _notification_message = (
            "Spark expectations job has been completed  \n\n"
            f"product_id: {self._context.product_id}\n"
            f"table_name: {self._context.get_table_name}\n"
            f"run_id: {self._context.get_run_id}\n"
            f"run_date: {self._context.get_run_date}\n"
            f"input_count: {self._context.get_input_count}\n"
            f"error_percentage: {self._context.get_error_percentage}\n"
            f"output_percentage: {self._context.get_output_percentage}\n"
            f"success_percentage: {self._context.get_success_percentage}\n"
            f"status: source_agg_dq_status = {self._context.get_source_agg_dq_status}\n"
            f"            source_query_dq_status = {self._context.get_source_query_dq_status}\n"
            f"            row_dq_status = {self._context.get_row_dq_status}\n"
            f"            final_agg_dq_status = {self._context.get_final_agg_dq_status}\n"
            f"            final_query_dq_status = {self._context.get_final_query_dq_status}\n"
            f"            run_status = {self._context.get_dq_run_status}"
        )

        _notification_hook.send_notification(
            _context=self._context, _config_args={"message": _notification_message}
        )

    def notify_on_failure(self, _error: str) -> None:
        """
        This function sends notification on failure of spark expectations project
        Args:
            _error: message or exception  for the failure(str)

        Returns: None

        """

        _notification_message = (
            "Spark expectations job has been failed  \n\n"
            f"product_id: {self._context.product_id}\n"
            f"table_name: {self._context.get_table_name}\n"
            f"run_id: {self._context.get_run_id}\n"
            f"run_date: {self._context.get_run_date}\n"
            f"input_count: {self._context.get_input_count}\n"
            f"error_percentage: {self._context.get_error_percentage}\n"
            f"output_percentage: {self._context.get_output_percentage}\n"
            f"status: source_agg_dq_status = {self._context.get_source_agg_dq_status}\n"
            f"            source_query_dq_status = {self._context.get_source_query_dq_status}\n"
            f"            row_dq_status = {self._context.get_row_dq_status}\n"
            f"            final_agg_dq_status = {self._context.get_final_agg_dq_status}\n"
            f"            final_query_dq_status = {self._context.get_final_query_dq_status}\n"
            f"            run_status = {self._context.get_dq_run_status}"
        )

        _notification_hook.send_notification(
            _context=self._context,
            _config_args={
                "message": _notification_message,
            },
        )

    def construct_message_for_each_rules(
        self,
        rule_name: str,
        failed_row_count: int,
        error_drop_percentage: float,
        set_error_drop_threshold: float,
        action: str,
    ) -> str:
        """
        This function supports constructing the notification message when rule threshold exceeds certain threshold
         Args:
            rule_name: name of the dq rule
            failed_row_count: number of failed of dq rule
            error_drop_percentage: error drop percentage
         Returns: str
        """

        _notification_message = (
            f"{rule_name} has been exceeded above the threshold "
            f"value({set_error_drop_threshold}%) for `row_data` quality validation\n"
            f"product_id: {self._context.product_id}\n"
            f"table_name: {self._context.get_table_name}\n"
            f"run_id: {self._context.get_run_id}\n"
            f"run_date: {self._context.get_run_date}\n"
            f"input_count: {self._context.get_input_count}\n"
            f"rule_name: {rule_name}\n"
            f"action: {action}\n"
            f"failed_row_count: {failed_row_count}\n"
            f"error_drop_percentage: {error_drop_percentage}\n\n\n"
        )

        return _notification_message

    def notify_on_exceeds_of_error_threshold_each_rules(
        self,
        message: str,
    ) -> None:
        """
        This function sends notification when specific rule error drop percentage exceeds above threshold
        Args:
            message: message to be sent in notification
        Returns: None

        """

        _notification_message = (
            f"Spark expectations - The number of notifications for rules being followed has surpassed "
            f"the specified threshold \n\n\n{message}"
        )

        _notification_hook.send_notification(
            _context=self._context, _config_args={"message": _notification_message}
        )

    def notify_rules_exceeds_threshold(self, rules: dict) -> None:
        """
        This functions identifies error drop percentage for rules which exceeds above set threshold
        Args:
            rules: lsit of rules which set to do data quality checks
        Returns:
            None

        """
        try:
            rules_failed_row_count: Dict[str, int] = {}
            notification_body = ""
            if self._context.get_summarised_row_dq_res is None:
                return None

            rules_failed_row_count = {
                itr["rule"]: int(itr["failed_row_count"])
                for itr in self._context.get_summarised_row_dq_res
            }

            for rule in rules[f"{self._context.get_row_dq_rule_type_name}_rules"]:
                if not rule["enable_error_drop_alert"]:
                    continue

                rule_name = rule["rule"]
                rule_action = rule["action_if_failed"]
                failed_row_count = int(
                    rules_failed_row_count[rule_name]
                    if rule_name in rules_failed_row_count
                    else 0
                )

                if failed_row_count is not None and failed_row_count > 0:
                    set_error_drop_threshold = int(rule["error_drop_threshold"])
                    error_drop_percentage = round(
                        (failed_row_count / self._context.get_input_count) * 100, 2
                    )

                    if error_drop_percentage >= set_error_drop_threshold:
                        notification_body = (
                            notification_body
                            + self.construct_message_for_each_rules(
                                rule_name=rule_name,
                                failed_row_count=failed_row_count,
                                error_drop_percentage=error_drop_percentage,
                                set_error_drop_threshold=set_error_drop_threshold,
                                action=rule_action,
                            )
                        )
                if notification_body != "":
                    self.notify_on_exceeds_of_error_threshold_each_rules(
                        notification_body
                    )

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"An error occurred while sending notification when the error threshold is breached: {e}"
            )



from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import Optional, Dict
import pluggy
from pyspark.sql.session import SparkSession
from spark_expectations import _log
from spark_expectations.config.user_config import Constants as UserConfig


SPARK_EXPECTATIONS_SECRETS_BACKEND = "spark_expectations_secrets_backend"

spark_expectations_secrets_plugin_spec = pluggy.HookspecMarker(
    SPARK_EXPECTATIONS_SECRETS_BACKEND
)


class SparkExpectationsSecretPluginSpec:
    @staticmethod
    @spark_expectations_secrets_plugin_spec(firstresult=True)
    def get_secret_value(
        secret_key_path: str, secret_dict: Dict[str, str]
    ) -> Optional["str"]:
        """Custom execute method that is able to be plugged in."""


@functools.lru_cache
def get_spark_expectations_tasks_hook() -> SparkExpectationsSecretPluginSpec:
    pm = pluggy.PluginManager(SPARK_EXPECTATIONS_SECRETS_BACKEND)
    pm.add_hookspecs(SparkExpectationsSecretPluginSpec)
    pm.load_setuptools_entrypoints(SPARK_EXPECTATIONS_SECRETS_BACKEND)
    pm.register(CerberusSparkExpectationsSecretPluginImpl())
    pm.register(DatabricksSecretsSparkExpectationsSecretPluginImpl())
    for name, plugin_instance in pm.list_name_plugin():
        _log.info(
            "Loaded plugin with name: %s and class: %s",
            name,
            plugin_instance.__class__.__name__,
        )
    return pm.hook


spark_expectations_secrets_backend_plugin_impl = pluggy.HookimplMarker(
    SPARK_EXPECTATIONS_SECRETS_BACKEND
)


class CerberusSparkExpectationsSecretPluginImpl(SparkExpectationsSecretPluginSpec):
    @staticmethod
    @spark_expectations_secrets_backend_plugin_impl
    def get_secret_value(
        secret_key_path: str, secret_dict: Dict[str, str]
    ) -> Optional[str]:
        """
        This function implemented to get secret value from cerberus
        Args:
            secret_key_path: str which accepts url with secret key
            secret_dict: dict which contains params for fetch secrets
        Returns:
            str | none : returns secret value in string or none
        """

        from cerberus.client import CerberusClient

        if secret_dict[UserConfig.secret_type].lower() == "cerberus":
            _client = CerberusClient(secret_dict[UserConfig.cbs_url])
            data = _client.get_secrets_data(secret_key_path)
            return data

        return None


class DatabricksSecretsSparkExpectationsSecretPluginImpl(
    SparkExpectationsSecretPluginSpec
):
    @staticmethod
    @spark_expectations_secrets_backend_plugin_impl
    def get_secret_value(
        secret_key_path: str, secret_dict: Dict[str, str]
    ) -> Optional[str]:
        """
         # pragma: no cover
        This function implemented to get secret value from databricks scope
        Args:
            secret_key_path: str which accepts url with secret key
            secret_dict: dict which contains params for fetch secrets
        Returns:
            str | none : returns secret value in string or none
        """
        if secret_dict[UserConfig.secret_type].lower() == "databricks":
            try:
                from pyspark.dbutils import DBUtils

                spark = SparkSession.getActiveSession()  # pragma: no cover
                dbutils = DBUtils(spark)  # pragma: no cover
            except ImportError:
                raise ImportError(
                    "You must install databricks to use the databricks secrets backend, "
                    "please try pip install databricks"
                )

            data = dbutils.secrets.get(
                scope=secret_dict[UserConfig.dbx_secret_scope], key=secret_key_path
            )  # pragma: no cover
            return data  # pragma: no cover
        return None  # pragma: no cover


@dataclass
class SparkExpectationsSecretsBackend:
    secret_dict: Dict[str, str]

    def get_secret(self, secret_key: Optional[str]) -> str | None:
        """
        This function implemented to get secret value
        Args:
            secret_key: Optional[str] which accepts url with secret key
        Returns:
            str | none : returns secret value in string or none
        """
        return get_spark_expectations_tasks_hook().get_secret_value(
            secret_key_path=secret_key, secret_dict=self.secret_dict
        )



import os
from pyspark.sql.session import SparkSession

current_dir = os.path.dirname(os.path.abspath(__file__))


def get_spark_session() -> SparkSession:
    if (
        os.environ.get("UNIT_TESTING_ENV")
        == "spark_expectations_unit_testing_on_github_actions"
    ) or (os.environ.get("SPARKEXPECTATIONS_ENV") == "local"):
        builder = (
            SparkSession.builder.config(
                "spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension"
            )
            .config("spark.jars.packages", "io.delta:delta-core_2.12:2.4.0")
            .config(
                "spark.sql.catalog.spark_catalog",
                "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            )
            .config("spark.sql.warehouse.dir", "/tmp/hive/warehouse")
            .config("spark.driver.extraJavaOptions", "-Dderby.system.home=/tmp/derby")
            .config("spark.jars.ivy", "/tmp/ivy2")
            .config(  # below jars are used only in the local env, not coupled with databricks or EMR
                "spark.jars",
                f"{current_dir}/../../jars/spark-sql-kafka-0-10_2.12-3.0.0.jar,"
                f"{current_dir}/../../jars/kafka-clients-3.0.0.jar,"
                f"{current_dir}/../../jars/commons-pool2-2.8.0.jar,"
                f"{current_dir}/../../jars/spark-token-provider-kafka-0-10_2.12-3.0.0.jar",
            )
        )
        return builder.getOrCreate()

    return SparkSession.getActiveSession()



from dataclasses import dataclass
from typing import Any, List, Dict
from pyspark.sql import DataFrame
from spark_expectations.core.exceptions import SparkExpectationsMiscException
from spark_expectations.sinks.utils.writer import SparkExpectationsWriter
from spark_expectations.core.context import SparkExpectationsContext


@dataclass
class SparkExpectationsCollectStatistics:
    """
    This class implements logging statistics on success and failure
    """

    _context: SparkExpectationsContext
    _writer: SparkExpectationsWriter

    def __post_init__(self) -> None:
        self.collect_stats_decorator = self.collect_stats_on_success_failure()

    def collect_stats_on_success_failure(self) -> Any:
        """
        The function implements decorator to log statistics on success and failure
        Returns:
            Any: function
        """

        def decorator(func: Any) -> Any:
            def wrapper(*args: List, **kwargs: Dict) -> DataFrame:
                try:
                    self._context.set_dq_start_time()

                    result = func(*args, **kwargs)

                    self._context.set_dq_run_status("Passed")
                    self._context.set_dq_end_time()

                    self._writer.write_error_stats()
                except Exception as e:
                    self._context.set_dq_run_status("Failed")
                    self._context.set_end_time_when_dq_job_fails()
                    self._context.set_dq_end_time()

                    self._writer.write_error_stats()
                    raise SparkExpectationsMiscException(e)
                return result

            return wrapper

        return decorator
